import time
import threading
from functools import wraps

# As ISTAT has put in place extreme restrictions (5 requests per minute otherwise the IP gets blacklisted for 7 days...),
# this decorator prevents that by tracking the number of requests and pausing them if needed. 

class RateLimiter:
    def __init__(self, max_calls, time_to_pass):
        self.call_count = 0
        self.last_reset_time = time.time()
        self.lock = threading.Lock()
        self.max_calls = max_calls
        self.time_to_pass = time_to_pass

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self.lock:
                if time.time() - self.last_reset_time > self.time_to_pass:
                    self.call_count = 0
                    self.last_reset_time = time.time()
                # Reset counter
                if self.call_count + 1 > self.max_calls:
                    print(f"{self.max_calls} requests limit reached. Waiting {self.time_to_pass} seconds...")
                    time.sleep(self.time_to_pass)
                    self.call_count = 0
                    self.last_reset_time = time.time()
                    print("Resuming work.")
                # Track the count
                self.call_count += 1
            # This is the wrapped function.
            return func(*args, **kwargs)
        return wrapper

rate_limiter = RateLimiter(max_calls=5, time_to_pass=70)


import pandas as pd
import requests
import xml.etree.ElementTree as ET
from .errors import DimensionsOrKwargsError, NotAListError, TooManyDimensionsError, TooManyDimensionsError2, DifferentDimensionValueError, KwargsError, OtherResponseCodeError, WrongFormatError
from .rate_limiter import rate_limiter
from datetime import datetime

@rate_limiter
def get_data(dataflow_id, timeout=60, dimensions=[], force_url=False, start_period="", end_period="", updated_after="", returned="dataframe", select_last_edition=True, debug_url=False, **kwargs):
    """
    

    Parameters
    ----------
    dataflow_id : String, 
        the dataflow id of the dataset.
    timeout : Int, 
        the maximum time before the request is aborted.
    dimensions : List, 
        an ordered list of strings of the dimensions. Make sure to leave it null if you use kwargs. The default is [].
    force_url : Bool, 
        used to force the URL request even if the they were not checked against the allowed dimensions. The default is False.
     debug_url : Bool, 
         used to obtain the generated URL for manual debugging. The default is False.
    start_period : Int, 
        used to filter for start period. The default is "".
    end_period : Int, 
        used to filter for end period. The default is "".
    updated_after : Int, 
        used to filter for update period. The default is "".
    returned : String, 
        "dataframe" or "csv", the format to be returned. The default is "dataframe".
    **kwargs : Key=value, 
        each kwarg will be used in place of the keys of the URL. Can't be used together with the dimensions list. Usage: freq="Q", correz="W"...


    Returns
    -------
    df : Returns a pandas DataFrame with all the dataflows if you choose the dataframe.
    csv file: Creates a csv file in the path of your code if you choose the csv.

    """
    def fetch_dimensions_df(timeout):
        nonlocal dimensions_df
        if dimensions_df.empty:
            dimensions_df = get_dimensions(dataflow_id, timeout)
        return dimensions_df

    dimensions = [string.upper() for string in dimensions]
    dimensions_df = pd.DataFrame() # Initialize to avoid using complex syntax to check if it exists
    if returned != "dataframe" and returned != "csv":
        raise WrongFormatError()
    if dimensions and kwargs:
        raise DimensionsOrKwargsError
    elif not dimensions and not kwargs:
        dimensions = ["all"]
    elif not isinstance(dimensions, list):
        raise NotAListError
        return None
    elif not force_url and dimensions:
        # Sometimes url checker can bug out for undiscovered reasons, in this case you are free to force the program to request data
        dimensions_df = fetch_dimensions_df(timeout)
        if len(dimensions) != (len(dimensions_df["dimension_id"].unique())):
            raise TooManyDimensionsError(dimensions, (len(dimensions_df["dimension_id"].unique())))
        
        counter = 0
        for _, user_dim in enumerate(dimensions):
            user_dim = user_dim.upper()
            if user_dim != "":
                counter_dim_df = dimensions_df[dimensions_df["order"] == counter+1]
                if user_dim in counter_dim_df["dimension_value"].tolist():
                    counter += 1
                else:
                    raise DifferentDimensionValueError(user_dim, counter_dim_df["dimension_id"].unique(), counter_dim_df["dimension_value"].tolist())
            else:
                counter += 1 
                
    elif kwargs: # The check must be done even if force_url==True as the program needs to fetch the positioning for dimension values.
        dimensions_df = fetch_dimensions_df(timeout)
        # Check how many dimensions there are
        for _ in range(len(dimensions_df["dimension_id"].unique())):
            dimensions.append("")
        for key, value in kwargs.items():
            check = False
            while not check:
                for index, row in dimensions_df.iterrows():
                    if key.casefold() == row["dimension_id"].casefold():
                        if value.casefold() == row["dimension_value"].casefold():
                            dimensions[row["order"]-1] = value # Order-1 is needed as order row starts from 1, order for lists starts from 0
                            check = True
                if check:
                    break
                raise KwargsError(key, value)
    
    # Important part: this feature allows users to always select the latest edition ISTAT has for the dataflow.
    # It uses the function find_last_edition to iter and find the last edition and automatically fills it for the user.
    # It won't work if an edition is manually added via kwargs or dimensions. It also checks whether the correct number of dimensions was added in order to gracefully fail.
    # There are multiple checks: 
    # Check if dimensions_df is already fetched from the controlo before to avoid another call;
    # Check if kwargs are used and t_bis is not defined;
    # Check if the dimensions are less than the order position for the dimensions.
    if select_last_edition==True:
        if dimensions_df.empty:
            dimensions_df = fetch_dimensions_df(timeout)
        last_edition = find_last_edition(dimensions_df)
        if last_edition is not None:
            filtered_df = dimensions_df.loc[dimensions_df['dimension_id'] == 'T_BIS', 'order']
            order_values = filtered_df.tolist()
            order = int(order_values[0])-1
            if kwargs and "t_bis" in kwargs:
                print("Warning: you passed an edition value 't_bis=x' with select_last_edition set on True. Remove the t_bis variable if you want the module to automatically fetch the last edition.")
                pass
            if order >= len(dimensions):
                raise TooManyDimensionsError2(dimensions, (len(dimensions_df["dimension_id"].unique())))
            if dimensions[order] != "":
                print("Warning: you passed an edition value for the edition in the dimensions provided, with select_last_edition set on True. Remove the t_bis variable if you want the module to automatically fetch the last edition.")
                pass
            else:
                try:
                    dimensions[order] = last_edition
                except TooManyDimensionsError as e:
                    print(e)
        else:
            print("Edition dimension not found. Skipping auto-fetching.")
    # Checking if time periods are formatted right and building the strings
    dim_string = '.'.join(dimensions)
    if start_period=="" and end_period=="":
        period_string = "all?"
    elif end_period=="":
        period_string = f"all?startPeriod={start_period}"
    elif start_period=="":
        period_string = f"all?endPeriod={end_period}"
    elif start_period=="" and end_period=="" and isinstance(start_period, int):
        period_string = f"all?startPeriod={start_period}&endPeriod={end_period}"
        period_string=""
    if updated_after != "":
        if period_string:
            period_string.append(f"&updatedAfter={updated_after}")
        else:
            period_string = f"all=updatedAfter={updated_after}"
    elif updated_after == "":
        easter_egg = "You've been blessed! Hello, pleased to meet you!"
    
    # Build the string and make the request: if the response is 200, then keep going.
    api_url = rf"https://esploradati.istat.it/SDMXWS/rest/data/{dataflow_id}/{dim_string}/{period_string}"
    if debug_url==True:
        print(api_url)
    response = requests.get(api_url, timeout=timeout)
    response_code = response.status_code
    if response_code != 200:
        raise OtherResponseCodeError(response_code)
    elif response.status_code == 200:
        response = response.content.decode('utf-8-sig')
        tree = ET.ElementTree(ET.fromstring(response)) 
        namespaces = {
            'message': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message',
            'generic': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/data/generic',
            'common': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common'
        }
        
        data = []
        for series in tree.findall('.//generic:Series', namespaces):
            series_key = {}
            series_key_element = series.find('generic:SeriesKey', namespaces)
            if series_key_element is not None:
                for value in series_key_element.findall('generic:Value', namespaces):
                    key_id = value.get('id')
                    value_text = value.get('value')
                    series_key[key_id] = value_text


            for obs in series.findall('generic:Obs', namespaces):
                obs_data = series_key.copy()
                obs_dimension = obs.find('generic:ObsDimension', namespaces)
                if obs_dimension is not None:
                    obs_data['TIME_PERIOD'] = obs_dimension.get('value')

                obs_value = obs.find('generic:ObsValue', namespaces)
                if obs_value is not None:
                    obs_data['OBS_VALUE'] = obs_value.get('value')

                data.append(obs_data)

        df = pd.DataFrame(data)

        if df.empty:
            print("No data retrieved. Open a request on GitHub, please.")
            return None
        else:
            if returned == "dataframe":
                return df
            elif returned == "csv":
                df.to_csv(f"{dataflow_id}_data_{datetime.now().strftime('%Y%m%d_%H%M%S')}.csv", index=False)


def find_last_edition(df):
    try:
        edition_df = df[df["dimension_id"] == "T_BIS"]
    except:
        return None

    edition_list = edition_df["dimension_value"].tolist()
    date_info = []

    for edition in edition_list:
        year = int(edition[0:4])
        month_start = edition.find("M") + 1
        month_end = month_start + 2 if edition[month_start:month_start + 2].isdigit() else month_start + 1
        month = int(edition[month_start:month_end])

        if "G" in edition:
            day_start = edition.find("G") + 1
            day_end = day_start + 2 if edition[day_start:day_start + 2].isdigit() else day_start + 1
            day = int(edition[day_start:day_end])
            date_part = f"{year}{str(month).zfill(2)}{str(day).zfill(2)}"
        else:
            day = ""
            date_part = f"{year}{str(month).zfill(2)}"

        suffix = ""
        if "_" in edition:
            suffix_pos = edition.find("_")
            suffix = edition[suffix_pos:]

        date_info.append((int(date_part), suffix))

    if not date_info:
        return None

    date_info.sort(reverse=True, key=lambda x: x[0])
    last_date, last_suffix = date_info[0]

    last_date_str = str(last_date)
    year = last_date_str[0:4]
    yearless = last_date_str[4:]

    if "G" in edition_list[0]:
        month = yearless[0:2] if yearless[0] != "0" else yearless[1:2]
        day = yearless[-2:] if yearless[-2] != "0" else yearless[-1]
        chosen_edition = f"{year}M{month}G{day}{last_suffix}"
    else:
        month = yearless[0:2] if yearless[0] != "0" else yearless[1:2]
        chosen_edition = f"{year}M{month}{last_suffix}"

    return chosen_edition

    
            
@rate_limiter
def get_dimensions(dataflow_id, timeout=60, lang="en", returned="dataframe", debug_url=False):
    """
    

    Parameters
    ----------
    dataflow_id : String, 
        the dataflow id of the dataset.
    lang : String, 
        "en" or "it", the language the search will be performed i n. The default is "en".
    get : Bool, 
        used only when called by the function get_dataframe() with force_url=False. The default is False.
    returned : String, 
        "dataframe" or "csv", the format to be returned. The default is "dataframe".
    debug_url: Bool, 
        Set to True if you want the URL the request is being sent to printed to the console.

    Returns
    -------
    df : Returns a pandas DataFrame with all the dataflows if you choose the dataframe.
    csv file: Creates a csv file in the path of your code if you choose the csv.

    """
    if "," in dataflow_id:        
        parts = dataflow_id.split(",")
        dataflow_id = parts[1]
        
    if returned != "dataframe" and returned != "csv":
        raise WrongFormatError()
    namespaces = {
        'message': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message',
        'structure': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure',
        'common': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common',
        'xml': 'http://www.w3.org/XML/1998/namespace'
    }
    data_url = f"https://esploradati.istat.it/SDMXWS/rest/availableconstraint/{dataflow_id}/?references=all&detail=full"
    if debug_url == True:
        print(data_url)

    response = requests.get(data_url, timeout=timeout)
    codelist_list = []
    response_code = response.status_code
    if response_code != 200:
        raise OtherResponseCodeError(response_code)
            
    response = response.content.decode('utf-8-sig')
    tree = ET.ElementTree(ET.fromstring(response))
    cube_region = tree.find('.//structure:CubeRegion', namespaces)
    key_values = cube_region.findall('.//common:KeyValue', namespaces)

    codelist_list = []

    for codelist in tree.findall(".//structure:Codelist", namespaces):
        codelist_id = codelist.get('id')[3:]  # The prefix "CL_" must be removed as it is internal
        codelist_name = codelist.find(f'.//common:Name[@xml:lang="{lang}"]', namespaces).text

        for code in codelist.findall('.//structure:Code', namespaces):
            code_id = code.get('id')
            code_name = code.find(f'.//common:Name[@xml:lang="{lang}"]', namespaces).text

            for idx, key_value in enumerate(key_values):
                for value in key_value.findall('common:Value', namespaces):
                    if value.text == code_id:
                        codelist_list.append({
                            'dimension_id': codelist_id,
                            'dimension_name': codelist_name,
                            'dimension_value': code_id,
                            'value_explanation': code_name,
                            'order': idx + 1
                        })
                        break
        
        
    df = pd.DataFrame(codelist_list)
    if returned == "dataframe":
        return df
    elif returned == "csv":
        df.to_csv(f"{dataflow_id}_dimensions")

        


import pandas as pd
import requests
import xml.etree.ElementTree as ET
from .errors import OtherResponseCodeError, WrongFormatError
from .rate_limiter import rate_limiter

all_dataflows = pd.DataFrame() # Cardinal sin: but I used a global variable. It serves ONLY to avoid repeating requests when searching is employed.

@rate_limiter
def get_all_dataflows(returned="dataframe", timeout=30):
    """
    This function is used in the search_dataflows function to search for dataflows,
    but it can also be used alone to get all the possible dataflows.

    Returns
    -------
    df : Returns a pandas DataFrame with all the dataflows if you choose the dataframe.
    csv file: Creates a csv file in the path of your code if you choose the csv.

    """
    # This is the ISTAT url for all dataflows
    global all_dataflows
    if all_dataflows.empty:
        dataflow_url = "https://esploradati.istat.it/SDMXWS/rest/dataflow/ALL/ALL/LATEST"   
        response = requests.get(dataflow_url, timeout=timeout)
        response_code = response.status_code
        if response_code == 200:
            response = response.content.decode('utf-8-sig')
            tree = ET.ElementTree(ET.fromstring(response))
            # Namespaces for ISTAT' SDMX dataflows
            namespaces = {
                'message': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message',
                'structure': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure',
                'common': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common'
            }
            data = []
            for dataflow in tree.findall('.//structure:Dataflow', namespaces):
    
                name_it = None
                name_en = None
                for name in dataflow.findall('.//common:Name', namespaces):
                    lang = name.get('{http://www.w3.org/XML/1998/namespace}lang')
                    if lang == 'it':
                        name_it = name.text
                    elif lang == 'en':
                        name_en = name.text
                row = {
                    'id': dataflow.get('id'),
                    'agencyID': dataflow.get('agencyID'),
                    'version': dataflow.get('version'),
                    'isFinal': dataflow.get('isFinal'),
                    'name_it': name_it,
                    'name_en': name_en
                }
                data.append(row)
        else:
            raise OtherResponseCodeError(response_code)
            return None
    else:
        df = all_dataflows
        
    df = pd.DataFrame(data)
    all_dataflows = df # Updates the global variable.
    
    if returned.casefold() == "dataframe" :  
        return df
    elif returned.casefold() == "csv":
        df.to_csv("all_dataflows_ISTAT.csv")
    else:
        raise WrongFormatError()
        
def search_dataflows(search_term, mode="fast", lang="en", returned="dataframe", timeout=30):
    """
    Allows searching for dataflows starting from strings passed. Can also accept a list.

    Parameters
    ----------
    search_term : String or list of strings, 
        is required to perform a search through the datasets.
    mode : String, 
        can be deep or fast. Deep search requires more requests but also gets the dimensions for datasets in a readable way. The default is "fast".
    lang : String, 
        "en" or "it", the language the search will be performed in. The default is "en".
    returned : String, 
        "dataframe" or "csv", the format to be returned. The default is "dataframe".
 
    Raises
    ------
    errors
        OtherResponseCodeError: when the code response from the API URL is not 200.

    Returns
    -------
    df : Returns a pandas DataFrame with all the dataflows if you choose the dataframe.
    csv file: Creates a csv file in the path of your code if you choose the csv.

    """
    global all_dataflows
    if returned != "dataframe" and returned != "csv":
        raise WrongFormatError()
    # The function must accept either single words or lists
    if isinstance(search_term, str):
        search_term = [search_term]
    if all_dataflows.empty:
        df = get_all_dataflows(timeout=timeout)
    else:
        df = all_dataflows
    if df.empty:
        print("Error: cannot retrieve dataflows from the ISTAT API. Open a request on Github.")
        return None
    
    # Initialize dataframe
    search_df = df.copy()
    search_df = search_df.iloc[:0]
    for term in search_term:
        if lang == "en":
            temp_df = df[df["name_en"].str.contains(term, case=False, na=False)]
            search_df = pd.concat([search_df, temp_df], ignore_index=True)
        elif lang == "it":
            temp_df = df[df["name_it"].str.contains(term, case=False, na=False)]
            search_df = pd.concat([search_df, temp_df], ignore_index=True)
        elif lang == "id":
            temp_df = df[df["id"].str.contains(term, case=False, na=False)]
            search_df = pd.concat([search_df, temp_df], ignore_index=True)
        else:
            print("Language not found.")
    if search_df.empty:
        print(f"Warning: the dataflow {term} could not be found. Did you set the right language?")
        return None
    if mode == "fast":
        if returned == "dataframe":
            return search_df
        elif returned == "csv":
            search_df.to_csv("requested_data.csv", index=False)
    if mode =="deep":
        deep_search_df = deep_search(search_df, timeout=timeout)
        if returned == "dataframe":
            return deep_search_df
        elif returned == "csv":
            deep_search_df.to_csv("requested_data.csv", index=False)
        
def format_dimensions(codelist_list): # Format dimensions in a different function to avoid cluttering, only used in deeps_search
    codelist_list = sorted(codelist_list, key=lambda x: x['order'])
    dimension_dict = {}    
    for item in codelist_list:
        dimension_name = item['dimension_name']
        dimension_value = item['dimension_value']
        if dimension_name not in dimension_dict:
            dimension_dict[dimension_name] = []
        if dimension_value not in dimension_dict[dimension_name]:
            dimension_dict[dimension_name].append(dimension_value)

    formatted_parts = [f"{name}={','.join(values)}" for name, values in dimension_dict.items()]
    return ";".join(formatted_parts)

@rate_limiter
def get_dimension_dataflows(dataflow_id, timeout=30):
    """
    This function is called by deep_search and is used to retrieve the data from ISTAT endpoint while tracking the number of requests.

    Parameters
    ----------
    dataflow_id : this is the id, provided by deep_search.

    Raises
    ------
    errors
        OtherResponseCodeError: when the code response from the API URL is not 200.

    Returns
    -------
    tree : the tree that will be parsed by deep_search.

    """
    data_url = f"https://esploradati.istat.it/SDMXWS/rest/availableconstraint/{dataflow_id}/?references=all&detail=full"
    response = requests.get(data_url)
    response_code = response.status_code
    if response_code == 200:
        response = response.content.decode('utf-8-sig')
    else:
        raise OtherResponseCodeError(response_code)
        return None
    tree = ET.ElementTree(ET.fromstring(response))
    return tree
    
def deep_search(df, lang="en", timeout=30):  
    """
    This function is used by the search_dataflows function if the selected mode is "deep".

    Parameters
    ----------
    df : Must be a DataFrame.
    lang : String, 
        used to select the language of the search. The default is "en".

    Raises
    ------
    errors
        OtherResponseCodeError: when the code response from the API URL is not 200.

    Returns
    -------
    df : normal return when used by search_dataflows.

    """
      
    namespaces = {
        'message': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/message',
        'structure': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/structure',
        'common': 'http://www.sdmx.org/resources/sdmxml/schemas/v2_1/common',
        'xml': 'http://www.w3.org/XML/1998/namespace'
    }
    codelist_list = []
    df["dataflow_dimensions"] = ""
    for index, row in df.iterrows():
        dataflow_id = row["id"]
        tree = get_dimension_dataflows(dataflow_id, timeout=timeout)
        if tree is None:
            print("Error: the dataflow id could not be retrieved. Aborting.")
            return None
        cube_region = tree.find('.//structure:CubeRegion', namespaces)
        key_values = cube_region.findall('.//common:KeyValue', namespaces)

        for codelist in tree.findall(".//structure:Codelist", namespaces):
            codelist_name = codelist.find(f'.//common:Name[@xml:lang="{lang}"]', namespaces).text

            for code in codelist.findall('.//structure:Code', namespaces):
                code_id = code.get('id')

                for idx, key_value in enumerate(key_values):
                    for value in key_value.findall('common:Value', namespaces):
                        if value.text == code_id:
                            codelist_list.append({
                                'dimension_name': codelist_name,
                                'dimension_value': code_id,
                                'order': idx + 1
                            })
                            break
        dict_list = format_dimensions(codelist_list)
        df.at[index, 'dataflow_dimensions'] = dict_list
        
    return df


df = search_dataflows("Coltivazion", lang="it")



# -*- coding: utf-8 -*-
"""
Created on Fri May 30 12:37:57 2025

@author: DiMartino
"""

class DimensionsOrKwargsError(Exception):
    """
    Warning: Either use a dimensions list (make sure it is ordered, get the correct
    order from get_dimensions) or use kwargs, never both.
    """
    def __init__(self, message="Warning: either pass the dimensions in the correct order as a list or use kwargs."):
        self.message = message
        super().__init__(self.message)
        
class NotAListError(Exception):
    """
    Warning: the variable Dimensions must be a list or null. Remember that if you provide a list,
          you must also provide wildcard values with a void string like ''.
    """
    def __init__(self, message="""Warning: the variable Dimensions must be a list or null. Remember that if you provide a list,
          you must also provide wildcard values with a void string like ''."""):
        self.message = message
        super().__init__(self.message)
        
class TooManyDimensionsError(Exception):
    """
    The dimensions you chose are less or more than those required from the dataflow.
    """
    def __init__(self, dimensions, dimensions_length, message=None):
        if message is None:
           message = f"""Warning: the dimensions you chose are {len(dimensions)}, while the dimensions
                 requested by the dataflow are {dimensions_length}. If you believe
                 this is a mistake, you can force the url by adding force_url=True to the function call."""
        self.message = message
        super().__init__(self.message)
        
class TooManyDimensionsError2(Exception):
    """
    The dimensions you chose are less or more than those required from the dataflow.
    """
    def __init__(self, dimensions, dimensions_length, message=None):
        if message is None:
           message = f"""Warning: the dimensions you chose are {len(dimensions)}, while the dimensions
                 requested by the dataflow are {dimensions_length}. The edition auto-fetch cannot work."""
        self.message = message
        super().__init__(self.message)
        
class DifferentDimensionValueError(Exception):
    """
    Different dimensions found.
    """
    def __init__(self, user_dim, dataflow_dim, dataflow_values, message=None):
        if message is None:
            message = f"""Warning: the dimension value {user_dim} cannot be found in the possible values for the dimension {dataflow_dim}. Check if the order of the dimensions you added is correct. 
Available values for the dimension: {dataflow_values}.
If you believe this is an error, you can force the url by adding force_url=True to the function call."""
        self.message = message
        super().__init__(self.message)
        
class KwargsError(Exception):
    """
    Error while using arguments. Check the name and value of arguments or use a list.
    """
    def __init__(self, key, value, message=None):
        if message is None:
            message = f"Error while using arguments. Check the name and value of arguments or use a list. {key}:{value} could not be found."
        self.message = message
        super().__init__(self.message)
        
class OtherResponseCodeError(Exception):
    """
    Different response code from 200 found.
    """
    def __init__(self, response_code, message=None):
        if message is None:
            message = f"""Error {response_code}. Check SDMX documentation and double check the dataflow id spelling."""
        self.message = message
        super().__init__(self.message)
        
        
class WrongFormatError(Exception):
    """
    Error while determining format.
    """
    def __init__(self, message="Wrong format requested. Choose either 'csv' or 'dataframe'."):
        self.message = message
        super().__init__(self.message)